/**

	.NET rendering layer for ElevaSim simulator

	Wrappers function for els_control

	Jan Raszyk, 2006-2007

*/

#include "els_control.h"
#include "els_renderer.h"
#include <fstream>
#include <iostream>
#include <iomanip>
#include <stdarg.h>
#include <string>
#include <sstream>

using namespace System;
using namespace System::Windows::Forms;
using namespace System::Threading;
using namespace System::Runtime::InteropServices;
using namespace System::Drawing;

namespace elevasim {

	// create singleton
	Renderer renderer;

	// converts .NET string into a STL std::string
	// I'd never expect this has to be so ugly
	std::string STR2STR(System::String ^ s)
	{
		System::IntPtr ptr = Marshal::StringToHGlobalAnsi(s);
		std::string std_s(static_cast<const char *>(static_cast<void *>(ptr))); 
		System::Runtime::InteropServices::Marshal::FreeCoTaskMem(ptr);
		return std_s;
	}

	Renderer::Renderer(void) {
		this->_speed = SSPDR1;
		this->_state = SIMSTOPPED;
		this->terminatesignal = true;
	}

	bool Renderer::SimInit(System::String ^ filename) {
		return this->r.SimInit(STR2STR(filename));
	}

	void Renderer::SimClear(void) {
		this->r.SimClear();
		Elapse(-1);
	}

	bool Renderer::DumpStats(System::String ^ filename)
	{
		return this->r.DumpStats(STR2STR(filename));
	}

	void Renderer::Elapse(int ms) {
		if (_speed == SSPDMAX) return;

		double elapsednext = seconds_elapsed + (ms / 1000.0) * _speed;

		while (r.TimeElapsed() < elapsednext) {
			if (!r.Tick()) {
				SimStop();
				break;
			}
		}

		seconds_elapsed = elapsednext;
	}

	bool Renderer::SimStart(void) {
		stop_at = 0;
		if (SimRunning()) return false;
		if (_speed == SSPDMAX) {
			// for max speed we need to run a new thread
			StartTicking();	
		}
		_state = SIMRUNNING;
		return true;
	}

	bool Renderer::SimRunFor(int t)
	{
		_speed = SSPDMAX;
		stop_at = (int) r.TimeElapsed() + t;
		StartTicking();
		_state = SIMRUNNING;
		return true;
	}

	void Renderer::SimStop(void) {
		if (!SimRunning()) return;
		if (_speed == SSPDMAX) {
			// send signal to ticking thread to terminate itself
			terminatesignal = true;
		}
		_state = SIMSTOPPED;
	}

	void Renderer::StatusInsert(std::string s)
	{
		this->statusbuffer += s;
	}

	System::String^ Renderer::StatusBufferFlush(void)
	{
		std::string buf = this->statusbuffer;
		this->statusbuffer.clear();
		return (gcnew System::String(buf.c_str()));
	}

	System::String^ Renderer::SimDetails(void)
	{
		std::stringstream ss;
		int seconds = (int) r.TimeElapsed();
		int days = seconds / (60*60*24) + 1;
		int hours = (seconds / (60*60)) % 24;
		int minutes = (seconds / 60) % 60;
		seconds = seconds % 60;

		ss << "Floors: " << r.FloorsCount() << std::endl;
		ss << "Lifts: " << r.LiftsCount() << std::endl;
		ss << "Day: " << days << "." << std::endl;
		ss << "Time: " << std::setw (2) << std::setfill ('0') << hours << ":" << std::setw (2) << std::setfill ('0') << minutes << ":" << std::setw (2) << std::setfill ('0') << seconds << std::endl;
		return (gcnew System::String(ss.str().c_str()));
	}

#define QUEUEWIDTH 150
#define WIDTHPERPERSON 4
#define FLOORASPECT 7
#define TOPMARGIN 10
#define LEFTMARGIN 10
#define LIFTHEIGHT 3*FLOORASPECT

	enum dots_align { da_rand, da_left, da_right };

	void Renderer::DrawPeople(Graphics ^g, int c, int x, int y, int w, int h, dots_align al)
	{
		int i;
		int px, py;
		const int eh = 8; const int ew = 8;

		Brush ^b = gcnew SolidBrush(Color::Orange);
		h -= eh; w -= ew; x += ew/2; y += eh/2;

		for (i = 0; i < c; i++)
		{
			switch (al) {
			case da_rand:
				px = x + ew * (i / 2);
				py = y + eh * (i % 2);
				break;

			case da_left:
				px = (int) (x + ew * i);
				py = y;
				break;
				
			default:
				px = (int) (x + w - ew * 1);
				py = y;
				break;

			}

			if (px > x+w) px = x+w;
			if (px < x) px = x;
			g->FillEllipse(b, px, py, ew, eh);
		}
	}

	void Renderer::DrawBuilding(Graphics ^g)
	{
		int buildingheight = (int) r.BuildingHeight()*FLOORASPECT;
		int lc = r.LiftsCount();
		int fc = r.FloorsCount();
		int px, py;
		int i;
		int sw; // lift shaft width
		int tsw; // total width of all lift shafts
		int buildingwidth = 2*QUEUEWIDTH;

		Pen ^p = gcnew Pen(Color::Black);
		Brush ^b = gcnew SolidBrush(Color::LightGreen);
		Brush ^b2 = gcnew SolidBrush(Color::LightBlue);
		Brush ^b3 = gcnew SolidBrush(Color::Gray);
		Font ^f = gcnew Font("Arial", 10);

		tsw = 0; for (i = 0; i < lc; i++) { tsw += r.LiftCapacity(i) * WIDTHPERPERSON; }
		tsw += WIDTHPERPERSON*lc;	// add padding to each lift
		buildingwidth += tsw;

		// building borders
		g->DrawRectangle(p, LEFTMARGIN, TOPMARGIN, buildingwidth, buildingheight);
		g->FillRectangle(b, LEFTMARGIN+1, TOPMARGIN+1, buildingwidth-1, buildingheight-1);
		
		px = LEFTMARGIN + QUEUEWIDTH;
		// lift shafts & boxes
		for (int i = 0; i < lc; i++)
		{
			sw = r.LiftCapacity(i) * WIDTHPERPERSON + WIDTHPERPERSON;
			py = TOPMARGIN + buildingheight - (int) r.LiftPosition(i)*FLOORASPECT - LIFTHEIGHT;
			g->FillRectangle(b3, px, TOPMARGIN, sw, buildingheight); 
			g->FillRectangle(b2, px, py, sw, LIFTHEIGHT);
			DrawPeople(g, r.PeopleInLift(i), px, py, sw, LIFTHEIGHT, da_rand);
			px += sw;
		}

		// floor 
		for (int i = 0; i < fc; i++)
		{
			py = TOPMARGIN + buildingheight - (int) r.FloorPosition(i)*FLOORASPECT;
			g->DrawLine(p, LEFTMARGIN, py, LEFTMARGIN+buildingwidth, py);
			DrawPeople(g, r.PeopleInFloorR(i), LEFTMARGIN, py-LIFTHEIGHT, QUEUEWIDTH, LIFTHEIGHT, da_left);
			DrawPeople(g, r.PeopleInFloorQU(i), LEFTMARGIN + QUEUEWIDTH + tsw, py-LIFTHEIGHT, QUEUEWIDTH, LIFTHEIGHT, da_left);
			DrawPeople(g, r.PeopleInFloorQD(i), LEFTMARGIN + QUEUEWIDTH + tsw, py-LIFTHEIGHT, QUEUEWIDTH, LIFTHEIGHT, da_right);
			g->DrawString(System::Int32(i).ToString(), f, b3, (float) LEFTMARGIN, (float) py - LIFTHEIGHT);
		}

	}

	void Renderer::DrawSim(System::Windows::Forms::PictureBox ^ pb)
	{
		if (!SimReady()) return;
			
		Bitmap ^ b = gcnew Bitmap(600, 500, Drawing::Imaging::PixelFormat::Format32bppRgb);
		
		DrawBuilding(Graphics::FromImage(b));

		pb->Image = b;
	}

	void Renderer::SkipSeconds(int s)
	{
		r.SkipSeconds(s);
		renderer.seconds_elapsed = renderer.r.TimeElapsed();
	}

	void KeepTicking() {
		renderer.terminatesignal = false;
		while(!renderer.terminatesignal && (!renderer.stop_at || (renderer.stop_at && renderer.r.TimeElapsed() < renderer.stop_at))) {
			if (!renderer.r.Tick()) {
				renderer.terminatesignal = true;
			}
		}
		renderer.SimStop();
		renderer.seconds_elapsed = renderer.r.TimeElapsed();
	}

	void StartTicking(void) {
		Thread tickthr(gcnew ThreadStart(KeepTicking));
		tickthr.Start();
	}

	// issues a termination signal to the max speed sim thread
	void StopTicking(void) {
		renderer.terminatesignal = true;
	}

	// used within the simulation core to output text description of actions happening
	void Con_Print(char* format, ...)
	{
		va_list argptr;
		static char string[2048];

		va_start (argptr, format);
		_vsnprintf_s(string, sizeof(string), format, argptr);
		renderer.StatusInsert(std::string(string));
		va_end (argptr);
	}
}
