/*
	
	ElevaSim, Jan Raszyk, 2006-2007

	Building class is defined here

*/

#include "els_common.h"

namespace elevasim {
	using namespace std;
	using namespace elevasim;

	Building::Building(void) {
		this->ticks = 0;
		this->elapsed_sec = 0;
	}

	void Building::InitOwnSpecs(istream & i, std::map<std::string, LiftSpecs> & lspecs, std::map<std::string, Floor> & fspecs)
	/* reads the specs for the building itself
	   is being called from the main input reading procedure */
	{
		string buf;
		map<int, FloorSpecs> floors_buf;

		while (i && (i >> buf, buf.compare(STRUCT_ENDING)))
		{

			if (!buf.compare("patra")) {
				FloorNums fn;
				std::set<int>::iterator sit;
				fn.Read(i);
				i >> buf;
				for (sit = fn.nums.begin(); sit != fn.nums.end(); sit++)
					floors_buf[*sit] = fspecs[buf];
				continue;
			}

			if (!buf.compare("vytah")) {
				FloorNums fn;
				fn.Read(i);
				i >> buf;
				this->_lifts.push_back(Lift(lspecs[buf]));
				this->_lifts.back().blding = this;
				this->_lifts.back().reachable_floors = fn;
				continue;
			}
		}

		for(map<int, FloorSpecs>::iterator it = floors_buf.begin(); it != floors_buf.end(); it++)
		{   // if the source map is not complete, a re-numbering will be done here, not safe operation!
			Floor f;
			f = it->second;
			f.floor_num = it->first;
			this->_floors.push_back(f);
			this->_floors.back().blding = this;
			if (this->_floors.size() - 1 != it->first)
				cerr << "Input error: Floor numbering mismatch (" << this->_floors.size() - 1 << " vs " << it->first << ")" << endl;
		}
	}

	void Building::Init_Floors(void)
	/* initializes all the preset floors (it is - adds residents to them) */
	{
		std::vector<Floor>::iterator mit;

		for (mit = this->_floors.begin(); mit != this->_floors.end(); mit++)
		{
			for (int i = 1; i <= mit->starting_residents; i++)
				mit->AddResident();
		}
	}

	void Building::Read(istream & i)
	/* Read specification from input stream */
	{
		// read buffer
		string buf;

		// templates
		std::map<std::string, LiftSpecs> lifts_specs;
		std::map<std::string, ResidentFlow> resident_flows;
		std::map<std::string, VisitorFlow> visitor_flows;
		std::map<std::string, Floor> floors_specs;

		while (i && (i >> buf, buf.compare(STRUCT_ENDING)))
		{
			if (!buf.compare("vytah")) {
				i >> buf;
				lifts_specs[buf.c_str()] = LiftSpecs(i);
				continue;
			}

			if (!buf.compare("obyvatele"))
			{
				i >> buf;
				ResidentFlow r;
				r.Read(i);
				resident_flows[buf.c_str()] = r;
				continue;
			}

			if (!buf.compare("navstevnici"))
			{
				i >> buf;
				VisitorFlow v;
				v.Read(i);
				visitor_flows[buf.c_str()] = v;
				continue;
			}

			if (!buf.compare("patro")) {
				i >> buf;
				FloorSpecs t;
				t.Read(i, resident_flows, visitor_flows);
				floors_specs[buf.c_str()] = t;
				continue;
			}

			if (!buf.compare("budova"))
			{
				this->InitOwnSpecs(i, lifts_specs, floors_specs);
				this->Init_Floors();
				continue;
			}

			if (!buf.compare("clear"))
			{
				// do nothing, because we erase this object in the outer main function
			}
		
		}
	}

	ostream & operator <<(ostream & o, Building & b)
	/* prints the state of the building */
	{
		o << "Floors," << (int) b._floors.size() << endl;

		o << "floor,residents,in queues" << endl;
		for(vector<Floor>::iterator f = b._floors.begin(); f != b._floors.end(); f++)
		{
			o << f->floor_num << "," << f->residents.size() << "," << f->lift_queue_down._queue.size() + f->lift_queue_up._queue.size() << endl;
		}
		o << "Total lifts," << (int) b._lifts.size() << endl;

		if (!b.trav_stats.empty()) {
			o << "Travel time,Waiting time" << endl;
			time_duration tt, wt; // travel_time, waiting_time
			time_duration mintt = b.trav_stats.front().trav_time;
			time_duration maxtt = 0;
			time_duration sumtt = 0;
			time_duration minwt = b.trav_stats.front().wait_time;
			time_duration maxwt = 0;
			time_duration sumwt = 0;
			int tc = 0;

			for (std::deque<trav_stats_t>::const_iterator ci = b.trav_stats.begin(); ci != b.trav_stats.end(); ci++)
			{
				tc++;
				tt = ci->trav_time;
				wt = ci->wait_time;
				o << tt << "," << wt << endl;
				if (tt > maxtt) maxtt = tt;
				if (tt < mintt) mintt = tt;
				sumtt += tt;
				if (wt > maxwt) maxwt = wt;
				if (wt < minwt) minwt = wt;
				sumwt += wt;
			}
			o << "Minimum Travel Time,Average Travel Time,Maximum Travel Time" << endl;
			o << mintt << "," << sumtt/tc << "," << maxtt << endl;
			o << "Minimum Wait Time,Average Wait Time,Maximum Wait Time" << endl;
			o << minwt << "," << sumwt/tc << "," << maxwt << endl;
		}

		o << "Lift, idle time, pending time, activity, stops" << endl;
		int i = 0;
		lift_stats_t ls;
		for (vector<Lift>::const_iterator li = b._lifts.begin(); li != b._lifts.end(); li++)
		{
			i++;
			ls = li->GetStats();
			o << i << ", " << ls.idle_time << ", " << ls.travel_time << ", " << ls.travel_time / (ls.travel_time + ls.idle_time) << ", " << ls.stops << endl;
		}
		return o;
	}

	bool Building::Tick(void)
	/* simulation clock tick */
	{
		this->ticks++;
		this->elapsed_sec += TICK_INTERVAL;

		for(vector<Lift>::iterator i = this->_lifts.begin(); i != this->_lifts.end(); i++)
			i->Tick();

		for(vector<Floor>::iterator f = this->_floors.begin(); f != this->_floors.end(); f++)
			f->Tick(((int) this->elapsed_sec / (60*60)) % 24);

		return true;
	}

	int Building::ChooseLift(int trav_from, int trav_to)
	{
		int i;
		int lc = _lifts.size();
		std::vector<int> av_lifts;

		for (i = 0; i < lc; i++)
		{
			FloorNums & fn = _lifts[i].reachable_floors;
			if (fn.nums.find(trav_from) != fn.nums.end() && fn.nums.find(trav_to) != fn.nums.end())
			{
				av_lifts.push_back(i);
			}
		}
		
		return av_lifts[rand() % av_lifts.size()];
	}

	void Building::LiftCall(int trav_from, int trav_to)
	{
		int ln = ChooseLift(trav_from, trav_to);

		if (trav_from < trav_to && _lifts[ln].specs_data.algorithm == COLLECT_BOTH)
			_lifts[ln].calls_up.insert(trav_from);
		else
			_lifts[ln].calls_down.insert(trav_from);

		_lifts[ln].NewTargetFloor();
		
		Con_Print("Floor %d: Calling lift %i (want travel to floor %d)\n", trav_from, ln, trav_to);
	}

	void Building::AddStats(Human * h)
	{
		trav_stats_t t;
		t.trav_time = h->duration_travel;
		t.wait_time = h->duration_idle;	
		this->trav_stats.push_back(t);
	}
}
