/**

	Common definitions for ElevaSim simulation core

	Jan Raszyk, 2006-2007

*/

#include <map>
#include <string>
#include <iostream>
#include <set>
#include <string>
#include <algorithm>
#include <vector>
#include <sstream>
#include <queue>
#include <cstdlib>
#include <ctime>

// keyword used to terminate structures in input files
#define STRUCT_ENDING "end"

// how many ticks does it give per one second
#define TICKS_PER_SECOND 2

// how long does one tick take
#define TICK_INTERVAL 0.5

// ammount of ticks to wait for one traveller to enter/leave the lift
#define WAIT_PER_TRAVELLER 1;

namespace elevasim {

	// wraps functions to control global simulation actions 
	// contains data pointers to all structures used within the simulation
	class Building;

	// represents one traveller, remembers where the person is travelling,
	// for how long, etc.
	class Human;

	// represents one floor in the building, contains waiting queues 
	// and idle persons in the building
	class Floor;

	// represents one lift in the building, lift algorithm is in here
	class Lift;

	// visitor flows specifications are stored in here
	class Flow;
	class Travel;

	typedef double time_duration;

	// a set of integers with an ability to be initialized from input stream
	// doesn't have to be necessairly used for floor numbers
	class FloorNums 
	{ 
	private:
		void InitVec(void);
	public:
		std::set<int> nums;
		std::vector<int> vec_nums;
		void Read(std::istream & i);
	};

	class Flow
	{
	public:
		int amount;
		void Read (std::istream & i);
	};

	class ResidentFlow : public Flow
	/* specs of a flow of residents (pair: how_many, where), abitility to init from istream*/
	{
	public:
		FloorNums target;
		void Read (std::istream & i);
	};

	class Travel
	/* full plan of the travel of a visitor */
	{
	public:
		std::vector<int> targets;
		std::vector<double> time_spent;
		void Read (std::istream & i);
	};

	class VisitorFlow : public Flow
	{
	public:
		Travel v_travel;
		void Read (std::istream & i);
	};

	// each human is being told what is happening to him
	enum HumanState { HS_IDLE, HS_WAIT, HS_TRAVEL };

	enum HumanType_t { HT_RESIDENT, HT_VISITOR };

	class Human
	{
	public:
		HumanType_t type;	// resident or visitor
		Travel my_flow;		// this visitor's flow
		int cur_stop;		// current stop number in the visitors travel
		time_duration cur_stop_time;	// how much time visitior spent on current stop
		Human();			// initialization of resident
		Human(const Travel &);	// initialization of visitor
		int target_floor;		// where am I travelling?
		int source_floor;		// from where am I travelling?
		void IdleTick(void);	// increases time counters
		void TravelTick(void);	// clock tick while travelling in a lift
		void WaitTick(void);	// clock tick while travelling in a lift
		bool FinalStop(void);	// is the visitor on it's final stop?
		time_duration duration_idle;	// for how long am I waiting in the queue?
		time_duration duration_travel;	// for how long am I travelling already?
	};

	typedef std::deque<Human*> QueueCont;
	typedef QueueCont::iterator QueueIter;

	// contains travellers waiting in a queue
	class HumanQueue
	{
	public:
		QueueCont _queue;
		void Tick(void);
		void IdleTick(Floor &);
	};

	// contains people idle somewhere in the building
	class HumanPile
	{
	private:
		std::vector<Human*> _pile;
	public:
		HumanPile();
		void Tick(int);
		int size(void) { return (int) this->_pile.size(); }
		void Push(Human*);
		Human* PopRandom(void);
	};

	enum FloorTickResult { FTR_NONE = 0, FTR_SENDMAN = 1, FTR_RECEIVEMAN = 2 };

	class Floor
	{
	private:
		int RandTo(int max);
		void AddToQueues(const ResidentFlow &);
		void AddVisitorsToQueues(const VisitorFlow &);

	public:
		double height;
		bool entrance;
		Building* blding;
		int starting_residents;
		std::map<int, ResidentFlow> resident_flow;
		std::map<int, VisitorFlow> visitor_flow;
		int floor_num;
		void AddResident(void);

		void Defaults(void);
		void Read(std::istream & i, std::map<std::string, ResidentFlow> & rflow, std::map<std::string, VisitorFlow> & vflow);
		HumanQueue lift_queue_up;
		HumanQueue lift_queue_down;
		HumanQueue visitors;
		HumanPile residents;
		void Tick(int hour);
	};

	typedef Floor FloorSpecsData;
	typedef Floor FloorSpecs;


	void Con_Print(char* format, ...);

	enum LiftAlgorithm { COLLECT_NONE, COLLECT_DOWN, COLLECT_BOTH };

	enum LiftMode { LIFT_GO, LIFT_BREAK, LIFT_ACCEL, LIFT_STAY };

	enum Direction { D_UP, D_DOWN };

	struct LiftSpecsData
	{
		double max_speed;
		int capacity;
		double duration_doors;
		double duration_wait;
		LiftAlgorithm algorithm;
	};

	class LiftSpecs : public LiftSpecsData
	{
	public:
		LiftSpecs(std::istream & i);
		LiftSpecs(void);
	};

	class Building;

	struct lift_stats_t {
		time_duration travel_time;
		time_duration idle_time;
		time_duration stops;
	};

	class Lift
	{
	private:
		lift_stats_t lift_stats;
		int Waiting(int addticks);
		void LoadFromQueue(QueueCont & q);
		void UnloadTravellers(void);
		void LoadTravellers(void);
		void MoveDown(int t_floor);
		void MoveUp(int t_floor);
		void StartMove(int t_floor);
		void DoMove(int t_floor);
		double TickDist(void);
		void TickTravellers(void);
		void Step(void);
	public:
		LiftSpecsData specs_data;

		Direction dir;
		LiftMode mode;
		Building* blding;
		FloorNums reachable_floors;

		std::multimap<int, Human*> travellers;
		int floor;
		int target_floor;
		double position; // relative to floors base level
		double speed;
		double acceleration; // not implemented yet

		// lift calls
		std::set<int> calls_up; 
		std::set<int> calls_down;

		Lift(const LiftSpecsData &);
		
		void NewTargetFloor(void);

		lift_stats_t GetStats(void) const;

		void Tick(void);
	};

	struct trav_stats_t {
		time_duration trav_time;
		time_duration wait_time;
	};

	class Building
	{
	private:
		std::vector<Lift> _lifts;
		std::vector<Floor> _floors;
		void Init_Floors(void);
		long ticks;
		double elapsed_sec;
		friend class Lift;
		friend class Floor;
		int ChooseLift(int, int);
		friend class Reader;
		friend std::ostream & operator << (std::ostream &, Building &);
		std::deque<trav_stats_t> trav_stats;

	public:
		Building();

		// controls reading of the input file
		void Read(std::istream & i);

		// reads details associated with the building itself, usually last part of the input file
		void InitOwnSpecs(std::istream & i, std::map<std::string, LiftSpecs> & lspecs, std::map<std::string, Floor> & fspecs);

		bool Tick(void);
		void StartSimulation(int simulation_duration);
		
		// calls a lift, alters floor waiting queues and lift stops
		void LiftCall(int trav_from, int trav_to);

		// called on the end of the travel, will read travellers experience and store it in the stats
		void AddStats(Human *);
	};

};
