/*

	Simulation controller
	elevasim, Jan Raszyk, 2006-2007

*/

#include "els_common.h"
#include "els_control.h"
#include <fstream>
#include <iostream>

namespace elevasim {

	Reader::Reader() {
		this->_b = new Building;
		this->_b_inited = false;
		this->_b_id = 1;
	}

	bool Reader::Tick()
	{
		if (!this->_b_inited) return false;

		return this->_b->Tick();
	}

	void Reader::SkipSeconds(int s)
	{
		if (!_b_inited) return;

		_b->elapsed_sec += s;
	}
	
	double Reader::BuildingHeight()
	{
		if (!this->_b_inited) return 0;

		double buf = 0;
		std::vector<Floor>::const_iterator ci;
		
		for (ci = this->_b->_floors.begin(); ci != _b->_floors.end(); ci++)
			buf += ci->height;
		
		return buf;
	}

	int Reader::FloorsCount(void) {
		if (!this->_b_inited) return 0;

		return this->_b->_floors.size();
	}

	int Reader::LiftsCount(void) {
		if (!this->_b_inited) return 0;

		return this->_b->_lifts.size();
	}

	double Reader::LiftPosition(int liftn)
	{
		if (liftn >= LiftsCount()) return 0;
		if (!this->_b_inited) return 0;

		return _b->_lifts[liftn].position + FloorPosition(_b->_lifts[liftn].floor);
	}

	int Reader::LiftCapacity(int liftn)
	{
		if (liftn >= LiftsCount()) return 0;

		return _b->_lifts[liftn].specs_data.capacity;
	}

	double Reader::FloorPosition(int floorn)
	{
		int fc = FloorsCount();
		static int lastcached_building_id = 0;
		static std::vector<double> floorpos_cache;
		
		if (floorn >= fc) return 0;

		if (_b_id != lastcached_building_id) // store in cache
		{	
			floorpos_cache.clear();
			floorpos_cache.push_back(0);	// first floor is on the bottom
			for (int i = 1; i < fc; i++)
				floorpos_cache.push_back(floorpos_cache[i-1] + _b->_floors[i-1].height);
			
			lastcached_building_id = _b_id;
			return FloorPosition(floorn);	// fetch from cache
		}
		else	// fetch from cache
		{
			return floorpos_cache[floorn];
		}
	}

	int Reader::PeopleInLift(int liftn)
	{
		if (liftn >= LiftsCount()) return 0;

		return _b->_lifts[liftn].travellers.size();
	}

	int Reader::PeopleInFloorQU(int floorn)
	{
		if (floorn >= FloorsCount()) return 0;
		return _b->_floors[floorn].lift_queue_up._queue.size();
	}

	int Reader::PeopleInFloorQD(int floorn)
	{
		if (floorn >= FloorsCount()) return 0;
		return _b->_floors[floorn].lift_queue_down._queue.size();
	}

	int Reader::PeopleInFloorR(int floorn)
	{
		if (floorn >= FloorsCount()) return 0;
		return _b->_floors[floorn].residents.size() + _b->_floors[floorn].visitors._queue.size();
	}


	double Reader::TimeElapsed(void) {
		if (!this->_b_inited) return 0;

		return this->_b->elapsed_sec;
	}

	bool Reader::SimReady(void) {
		return this->_b_inited;
	}

	bool Reader::SimInit(std::string fn) {
		if (!this->_b) return false;

		std::ifstream b(fn.c_str());
		if (!b.is_open()) return false;

		this->_b->Read(b);
		this->_b_inited = true;
		return true;
	}

	void Reader::SimClear(void) {
		if (!this->_b_inited) return;
		
		this->_b_inited = false;
		delete this->_b;
		this->_b = new Building;
		this->_b_id++;
	}

	bool Reader::DumpStats(std::string fn)
	{
		if (!this->_b) return false;

		std::ofstream b(fn.c_str());
		if (!b.is_open()) return false;
		
		b << *_b;

		return true;
	}
}
