/*
	
	ElevaSim, Jan Raszyk, 2006-2007

	Floors class is defined here

*/

#include "els_common.h"

using namespace std;

namespace elevasim {

	int ReadNum(istream & i)
	/* Reads positive integer from input, all other characters will be kept in the input */
	{
		int ret = 0;
		char c = 0;

		while (i && i.get(c) && c == ' ');
		
		if (c)
			i.putback(c);

		while (i && i.get(c))
		{
			if (c <= '9' && c >= '0')
				ret = ret * 10 + (c - '0');
			else
			{
				i.putback(c);
				break;
			}
		}
		return ret;
	}

	void FloorNums::InitVec(void)
	{
		std::set<int>::iterator i;
		for (i = this->nums.begin(); i != this->nums.end(); i++)
			this->vec_nums.push_back(*i);
	}

	void FloorNums::Read(istream & i)
	/* converts the input e.g. "1,3-5,9" to a set of {1,3,4,5,9} */
	{
		int intbuf1, intbuf2, j;
		char charbuf = '-';

		while (i && charbuf != ' ' && charbuf != '\n' && charbuf != '\t')
		{
			intbuf1 = ReadNum(i);
			if (!i)
			{
				this->nums.insert(intbuf1);
				break;
			}
			i.get(charbuf);
			switch (charbuf)
			{
			case '-':
				if (i)
				{
					intbuf2 = ReadNum(i);
					for (j = intbuf1; j <= intbuf2; j++)
						this->nums.insert(j);
				}	
				else
				{
					this->nums.insert(intbuf1);
					break;
				}
				if (i) 
					i.get(charbuf);
				break;
			case ',':
				this->nums.insert(intbuf1);
				break;
			default:
				this->nums.insert(intbuf1);
				break;
			}
		}

		this->InitVec();
	}

	void ResidentFlow::Read(istream & i)
	/* reads specification of resident flows */
	{
		string buf;

		while (i && (i >> buf, buf.compare(STRUCT_ENDING)))
		{
			if (!buf.compare("mnozstvi"))
				i >> this->amount;

			if (!buf.compare("kam"))
				this->target.Read(i);
		}
	}

	void Travel::Read(istream & i)
	/* reads a sequence of pairs floor/duration which determines the travel of a visitor */
	{
		string buf;
		int n1;
		double n2;

		while(i && i >> buf && buf.compare(STRUCT_ENDING))
		{
			n1 = atoi(buf.c_str());
			i >> buf;
			if (buf.compare(STRUCT_ENDING))
			{
				n2 = atof(buf.c_str());
				this->targets.push_back(n1);
				this->time_spent.push_back(n2);
			} else break;
		}
	}

	void VisitorFlow::Read(istream & i)
	/* reads the 'navstevnici' struct specs */
	{
		string buf;

		while (i && (i >> buf, buf.compare(STRUCT_ENDING)))
		{
			if (!buf.compare("mnozstvi"))
				i >> this->amount;

			if (!buf.compare("cesta"))
				this->v_travel.Read(i);
		}
	}

	void Floor::Defaults(void)
	{
		this->entrance = false;
		this->height = 3;
	}

	void Floor::AddResident(void)
	/* used in floor initialization only! */
	{
		Human* h = new Human;
		this->residents.Push(h);
	}

	void Floor::Read(std::istream & i, std::map<std::string, ResidentFlow> & rflow, std::map<std::string, VisitorFlow> & vflow)
	/* read from input stream */
	{
		string buf;
		this->Defaults();
		this->height = 0;
		this->entrance = false;
		this->starting_residents = 0;

		while(i && (i >> buf, buf.compare(STRUCT_ENDING)))
		{
			if (!buf.compare("vyska")) 
				i >> this->height;

			if (!buf.compare("vchod"))
				i >> this->entrance;

			if (!buf.compare("tok"))
			{
				FloorNums bufnums;
				ResidentFlow r;
				VisitorFlow v;
				bool t;

				bufnums.Read(i);
				i >> buf;

				if (rflow.find(buf) != rflow.end())
				{
					r = rflow[buf];
					t = true;
				}
				else if (vflow.find(buf) != vflow.end())
				{
					v = vflow[buf];
					t = false;
				}
				else
					continue;
				
				for (std::set<int>::iterator j = bufnums.nums.begin(); j != bufnums.nums.end(); j++)
				{
					if (t)
						this->resident_flow[*j] = r;
					else
						this->visitor_flow[*j] = v;
				}
			}

			if (!buf.compare("pocatek"))
				i >> this->starting_residents;
		}
	}

	int Floor::RandTo(int max)
	{   // returns an integer from range <0, 1, 2, ..., max>
		return (int) (((double) max + 1) * rand() / (RAND_MAX + 1.0));
		// too paranoid for return rand() % (max + 1);
	}

	void Floor::AddToQueues(const ResidentFlow & rf)
	// picks humans from residents and adds them to appropriate lift waiting queues
	{
		double prob = (double) rf.amount / (60.0 * 60 * TICKS_PER_SECOND);
		double rprob = (double) rand() / (double) RAND_MAX;
		int t_floor;
		int t_floors = (int) rf.target.vec_nums.size() - 1;
		Human* th;

		while (rprob < prob && this->residents.size()) // random number generator has chosen to generate a traffic in this tick
		{
			rprob++; // next time it should always be less then 0 unless we have to generate more than one traveller per tick
			t_floor = rf.target.vec_nums[this->RandTo(t_floors)];  // target floor number
			th = this->residents.PopRandom();					   // pointer to human who will travel
			th->source_floor = this->floor_num;					   // remeber where he comes from
			th->target_floor = t_floor;							   // tell him where he wants to travel; todoB: path-finding algor. to choose semi-target floor
			th->duration_idle = 0;
			th->duration_travel = 0;
			
			// call the lift for him
			blding->LiftCall(floor_num, t_floor);

			// place him into the queue
			if (t_floor > this->floor_num)
				this->lift_queue_up._queue.push_back(th);
			else
				this->lift_queue_down._queue.push_back(th);
		}
	}

	void Floor::AddVisitorsToQueues(const VisitorFlow & vf)
	// generates visitors on the floor
	{
		// 1.0 here means "generate all visitors in one minute"
		double prob = (double) vf.amount / (60.0 * 1.0 * TICKS_PER_SECOND);
		double rprob = (double) rand() / (double) RAND_MAX;
		Human* th;
		int t_fl;

		if (vf.v_travel.targets.empty()) return;

		while (rprob < prob)
		{
			rprob++;						// 0 in most cases, unless we are generating more travellers in one tick
			th = new Human(vf.v_travel);	// create a new human and give him his travel plan
			t_fl = vf.v_travel.targets[0];	// where he travels? first entry in the travel plan
			th->target_floor = t_fl;
			th->source_floor = floor_num;	// he comes from here
			th->duration_idle = 0;
			th->duration_travel = 0;
			
			blding->LiftCall(floor_num, t_fl);

			if (t_fl > this->floor_num)
				lift_queue_up._queue.push_back(th);
			else
				lift_queue_down._queue.push_back(th);
		}
	}

	void Floor::Tick(int hour)
	{	
		this->lift_queue_down.Tick();
		this->lift_queue_up.Tick();
		this->visitors.IdleTick(*this);

		if (!(this->blding->ticks % 100))
		// refresh calls for all waiting people in queues, sometimes their calls can get lost
		// e.g.: lift arrives but is full and leaves and the call gets cancelled
		{
			QueueIter i;
			if (!this->lift_queue_down._queue.empty())
				blding->LiftCall(floor_num, lift_queue_down._queue.front()->target_floor);

			if (!lift_queue_up._queue.empty())
				blding->LiftCall(floor_num, lift_queue_up._queue.front()->target_floor);
		}
		
		std::map<int, ResidentFlow>::iterator i = this->resident_flow.find(hour);

		if (i != resident_flow.end() && residents.size()) 
			AddToQueues(i->second);

		std::map<int, VisitorFlow>::iterator j = visitor_flow.find(hour);

		// condition says: generate travellers only in first minute of the given hour
		if ((j != visitor_flow.end()) && (((int) blding->elapsed_sec / 60) % 60) < 1)
			AddVisitorsToQueues(j->second);
	}

}
