/*
	
	ElevaSim, Jan Raszyk, 2006-2007

	Lifts class is defined here

*/

#include "els_common.h"

using namespace std;

namespace elevasim {

	LiftSpecs::LiftSpecs(void)
	{
		this->max_speed = 1;
	}

	LiftSpecs::LiftSpecs(istream & i)
	{
		string buf;

		while(i && (i >> buf, buf.compare(STRUCT_ENDING)))
		{
			if (!buf.compare("rychlost")) 
				i >> this->max_speed;
			
			if (!buf.compare("kapacita"))
				i >> this->capacity;

			if (!buf.compare("sber")) 
			{
				i >> buf;
				if (!buf.compare("dolu"))
					this->algorithm = COLLECT_DOWN;
				if (!buf.compare("obema"))
					this->algorithm = COLLECT_BOTH;
				if (!buf.compare("zadny"))
					this->algorithm = COLLECT_NONE;
			}
		}
	}

	Lift::Lift(const LiftSpecsData & s_data)
	{
		this->specs_data = s_data;
		this->acceleration = 0;
		this->dir = D_UP;
		this->mode = LIFT_STAY;
		this->floor = 0;
		this->target_floor = 0;
		this->speed = this->specs_data.max_speed;  // speed was first intended as an indicator for current speed but that was changed later
		this->lift_stats.idle_time = 0;
		this->lift_stats.travel_time = 0;
		this->lift_stats.stops = 0;
		this->position = 0;
	}

	// called when some new call has been added, will determine where the lift should go
	void Lift::NewTargetFloor(void)
	{
		if ((!calls_up.size() && !calls_down.size()) || this->mode == LIFT_STAY)
		{
			// no calls -> just stop on the closest floor
			target_floor = floor;
			return;
		}

		std::set<int>::iterator it;
		// lower_bound returns an iterator to the first element greater or equal to given value
		// upper_bound returns an iterator to the first element greater or equal to given value


		if (this->dir == D_UP)
		{
			if (calls_up.size())
			{
				it = this->calls_up.upper_bound(this->floor);

				if (it == this->calls_up.end())
					target_floor = *(--it); // we are still going up even though there's no reason for it -> we stop on the next floor
				else
					target_floor = *it;
			}
			else target_floor = *--calls_down.end();
		}
		else // this->dir == D_DOWN
		{
			if (calls_down.size())
			{
				it = this->calls_down.upper_bound(this->floor);

				if (it == this->calls_down.begin())
					target_floor = *it; // we are going down but we should go up! so at first we need to stop on this floor
				else
					target_floor = *(--it); // this number will be equal or lower to the current floor number
			}
			else target_floor = *calls_up.begin();
		} 
	}

	int Lift::Waiting(int addticks)
	{	// inner lift clock .. if you call it with parameter 20, it will return true for 20 times when called with -1
		static int a = 0; // buffer

		a += addticks;
		if (a < 0)
			a = 0;

		return a > 0;
	}

	void Lift::UnloadTravellers(void)
	// ensure travellers leave the lift and enter the building in appropriate place
	{
		typedef std::multimap<int, Human*>::iterator it;
		pair<it, it> range;
		it i;
		range = this->travellers.equal_range(this->floor);
		
		for (i = range.first; i != range.second; i++)
		{
			i->second->cur_stop_time = 0;
			this->blding->AddStats(i->second); // write stats
			if (i->second->type == HT_RESIDENT) 
			{
				// leave him alive!
				this->blding->_floors[this->floor].residents.Push(i->second); 
			}
			// fixme: we ignore all other path points because of wrong data structure usage somewhere else
			else // if (i->second->FinalStop())
			{
				// shouldn't go anywhere else
				delete i->second;	
			}
/*
			else
			{
				this->blding->_floors[floor].visitors._queue.push_back(i->second);
			}
*/
			Con_Print("Floor %d: Human unloaded (from floor %d)\n", this->floor, i->second->source_floor);
			this->Waiting(4);
		}

		this->travellers.erase(this->floor);
	}

	void Lift::LoadFromQueue(QueueCont & q)
	// loads humans from given queue into the lift
	{
		QueueCont::iterator i;
		int space_available = this->specs_data.capacity - (int) this->travellers.size();
		int where_going;
		Human* h;
		QueueCont::const_iterator qi;

		while (space_available && q.size())
		{
			h = q.front();
			Con_Print("Floor %d: Human loaded (travelling to floor %d)\n", this->floor, h->target_floor);
			this->Waiting(10);
			q.pop_front();
			where_going = h->target_floor;
			this->travellers.insert(make_pair(where_going, h));
			if (where_going > floor) calls_up.insert(where_going); else calls_down.insert(where_going);
			space_available--;
			this->Waiting(1);
		}

		NewTargetFloor();
	}

	void Lift::LoadTravellers(void)
	/* move people from floor inside the lift
	   pick-up only those people, who travel in the same direction as the lift does */
	{
		Floor & f = this->blding->_floors[this->floor];
		bool lift_is_free = calls_down.size() == 0 && calls_up.size() == 0;

		if (this->dir == D_DOWN || target_floor <= floor)
		{
			this->LoadFromQueue(f.lift_queue_down._queue);
		}

		if (this->dir == D_UP || target_floor >= floor)
		{
			this->LoadFromQueue(f.lift_queue_up._queue);
		}
	}

	double Lift::TickDist(void)
	{   // return the distance the lift should travel in full speed per one tick
		static double tickd = 0;
		if (!tickd)
			tickd = this->speed * TICK_INTERVAL; // m/s * s -> m

		return tickd;
	}

	void Lift::MoveDown(int t_floor)
	{   // moves the lift down for one tick distance
		this->dir = D_DOWN;
		this->position -= this->TickDist();
		if (this->position < 0)
		{
			if (this->floor == t_floor)
			{
				this->position = 0;
				this->mode = LIFT_STAY;
			}
			else
			{
				this->floor--;
				this->position = this->blding->_floors[this->floor].height + this->position;
			}
		}
	}

	void Lift::MoveUp(int t_floor)
	{   // moves the lift up for one tick distance
		double tf_height = this->blding->_floors[this->floor].height; // this floor's height
		this->dir = D_UP;
		this->position += this->TickDist();

		if (this->position >= tf_height)
		{
			this->floor++;
			if (this->floor == t_floor)
			{
				this->position = 0;
				this->mode = LIFT_STAY;
			}
			else
			{
				this->position -= tf_height;
			}
			if (floor >= (int) blding->_floors.size())
			{
				Con_Print("Wrong floor!\n");
			}
		}
			
	}

	void Lift::DoMove(int t_floor)
	{   // continues in moving the lift in the direction it was going before
		if (this->dir == D_DOWN && t_floor <= floor)
		{
			this->MoveDown(t_floor);
		}
		else if (t_floor > floor)
		{
			this->MoveUp(t_floor);
		}
		else
		{
			this->mode = LIFT_STAY;
		}
	}

	void Lift::StartMove(int t_floor)
	{   // moves the lift in a direction based on the floor we want to visit
		this->mode = LIFT_GO;
		if (this->floor > t_floor)
			this->MoveDown(t_floor);
		else if (this->floor < t_floor)
			this->MoveUp(t_floor);
		else
			this->mode = LIFT_STAY;
	}

	void Lift::TickTravellers(void)
	{   // send clock tick to all humans inside of the lift
		std::multimap<int, Human*>::iterator it;
		for (it = this->travellers.begin(); it != this->travellers.end(); it++)
			(*it).second->TravelTick();
	}

	void Lift::Step(void)
	{   // this is a main lift algorithm deciding what should the lift do next
		bool wasnt_there = true; // the lift wasn't in it's target floor when the tick began

		//Con_Print("Target floor: %i\n", target_floor);
		
		if (mode == LIFT_STAY)
		{
			wasnt_there = false;

			if (!Waiting(-1))
			{
				if (!calls_down.empty()) // enough of waiting, let's go!
				{
					calls_down.erase(target_floor);
					if (specs_data.algorithm == COLLECT_DOWN) calls_up.erase(target_floor);
					if (!calls_down.empty()) {
						mode = LIFT_GO;
						NewTargetFloor();
						StartMove(target_floor);
						if (mode == LIFT_STAY) wasnt_there = true;
					} else	// this happens when the lift was on the floor idle and somebody wants to enter it now
						wasnt_there = true;
				}
				else if (!calls_up.empty())
				{
					calls_up.erase(target_floor);
					if (!calls_up.empty()) {
						mode = LIFT_GO;
						NewTargetFloor();
						StartMove(target_floor);
						if (mode == LIFT_STAY) wasnt_there = true;
					} else	// this happens when the lift was on the floor idle and somebody wants to enter it now
						wasnt_there = true;
				}
			}
		}
		else // we are travelling somewhere, so just move forward
		{
			DoMove(target_floor);
		}

		if (wasnt_there && mode == LIFT_STAY)
		// we've just arrived to the target floor!
		{
			lift_stats.stops++;
			Con_Print("Floor %d: Lift arrival\n", floor);
			if (dir == D_DOWN) calls_down.erase(floor); else calls_up.erase(floor);
			UnloadTravellers();
			LoadTravellers();
		}
	}

	void Lift::Tick(void)
	{
		this->TickTravellers();
		this->Step();

		// update stats
		if (mode == LIFT_STAY) 
			lift_stats.idle_time += TICK_INTERVAL; 
		else
			lift_stats.travel_time += TICK_INTERVAL;
	}

	lift_stats_t Lift::GetStats(void) const
	{
		return lift_stats;
	}
}
